"""Streaming JSONL/NDJSON parser."""

import json
import os
from pathlib import Path
from typing import Generator, Optional

from .types import ColumnInfo, ParsedDataset

_SCHEMA_SAMPLE_LINES = 100


def _detect_encoding(path: str) -> str:
    """Detect file encoding using charset-normalizer."""
    try:
        from charset_normalizer import from_path
        result = from_path(path, cp_isolation=["utf-8", "latin-1", "cp1252"])
        if result.best():
            return str(result.best().encoding)
    except Exception:
        pass
    return "utf-8"


def _discover_columns(path: str, encoding: str) -> list:
    """Read first N lines to build ordered column list (union of all keys)."""
    columns: dict = {}
    with open(path, encoding=encoding, errors="replace") as f:
        for i, line in enumerate(f):
            line = line.strip()
            if not line:
                continue
            try:
                obj = json.loads(line)
                if isinstance(obj, dict):
                    for key in obj:
                        columns[key] = None
            except json.JSONDecodeError:
                continue
            if i >= _SCHEMA_SAMPLE_LINES:
                break
    return list(columns.keys())


def _row_generator(path: str, encoding: str, column_names: list) -> Generator:
    """Yield data rows as lists of strings, aligned to column_names."""
    with open(path, encoding=encoding, errors="replace") as f:
        for line in f:
            line = line.strip()
            if not line:
                continue
            try:
                obj = json.loads(line)
            except json.JSONDecodeError:
                continue
            if not isinstance(obj, dict):
                continue
            yield [
                str(obj[col]) if col in obj and obj[col] is not None else ""
                for col in column_names
            ]


def parse_jsonl(
    path: str,
    encoding: Optional[str] = None,
) -> ParsedDataset:
    """Parse a JSONL/NDJSON file and return a streaming ParsedDataset."""
    path = str(Path(path).resolve())
    file_size = os.path.getsize(path)

    if not encoding:
        encoding = _detect_encoding(path)

    column_names = _discover_columns(path, encoding)
    if not column_names:
        raise ValueError(f"No valid JSON objects found in {path}")

    columns = [ColumnInfo(name=name, position=i) for i, name in enumerate(column_names)]

    with open(path, encoding=encoding, errors="replace") as f:
        sample = f.read(65536)
    sample_lines = sample.count("\n")
    if sample_lines > 0:
        bytes_per_line = len(sample.encode(encoding, errors="replace")) / sample_lines
        estimated_rows = int(file_size / bytes_per_line)
    else:
        estimated_rows = 0

    metadata = {
        "encoding": encoding,
        "delimiter": None,
        "header_row": None,
        "estimated_rows": estimated_rows,
        "file_size": file_size,
    }

    return ParsedDataset(
        columns=columns,
        row_iterator=_row_generator(path, encoding, column_names),
        metadata=metadata,
    )



"""Tests for Parquet parser and index_local integration."""

import pytest

from jdatamunch_mcp.tools.index_local import index_local
from jdatamunch_mcp.storage.data_store import DataStore

pytestmark = pytest.mark.skipif(
    not pytest.importorskip("pyarrow", reason="pyarrow not installed"),
    reason="pyarrow not installed",
)


def test_parse_parquet_columns(sample_parquet):
    from jdatamunch_mcp.parser.parquet_parser import parse_parquet
    parsed = parse_parquet(sample_parquet)
    assert len(parsed.columns) == 5
    assert [c.name for c in parsed.columns] == ["id", "name", "age", "city", "score"]


def test_parse_parquet_rows(sample_parquet):
    from jdatamunch_mcp.parser.parquet_parser import parse_parquet
    parsed = parse_parquet(sample_parquet)
    rows = list(parsed.row_iterator)
    assert len(rows) == 10
    assert rows[0][1] == "Alice"
    assert rows[0][0] == "1"


def test_parse_parquet_null_handling(sample_parquet):
    from jdatamunch_mcp.parser.parquet_parser import parse_parquet
    parsed = parse_parquet(sample_parquet)
    rows = list(parsed.row_iterator)
    frank = rows[5]
    assert frank[2] == ""   # age is None
    assert frank[4] == ""   # score is None


def test_parse_parquet_metadata(sample_parquet):
    from jdatamunch_mcp.parser.parquet_parser import parse_parquet
    parsed = parse_parquet(sample_parquet)
    assert parsed.metadata["estimated_rows"] == 10
    assert parsed.metadata["file_size"] > 0


def test_index_parquet(sample_parquet, storage_dir):
    result = index_local(path=sample_parquet, name="sample-parquet", storage_path=storage_dir)
    assert "error" not in result
    r = result["result"]
    assert r["rows"] == 10
    assert r["columns"] == 5
    assert r.get("skipped") is not True


def test_index_parquet_column_types(sample_parquet, storage_dir):
    index_local(path=sample_parquet, name="sample-parquet", storage_path=storage_dir)
    store = DataStore(base_path=storage_dir)
    idx = store.load("sample-parquet")
    col_types = {c["name"]: c["type"] for c in idx.columns}
    assert col_types["id"] == "integer"
    assert col_types["name"] == "string"
    assert col_types["score"] == "float"


def test_parquet_incremental(sample_parquet, storage_dir):
    index_local(path=sample_parquet, name="sample-parquet", storage_path=storage_dir)
    result2 = index_local(
        path=sample_parquet, name="sample-parquet", incremental=True, storage_path=storage_dir
    )
    assert result2["result"].get("skipped") is True


def test_parquet_missing_pyarrow(tmp_path, storage_dir, monkeypatch):
    """Graceful error when pyarrow is not installed."""
    import builtins
    real_import = builtins.__import__

    def mock_import(name, *args, **kwargs):
        if name == "pyarrow" or name == "pyarrow.parquet":
            raise ImportError("No module named 'pyarrow'")
        return real_import(name, *args, **kwargs)

    p = tmp_path / "fake.parquet"
    p.write_bytes(b"PAR1")

    monkeypatch.setattr(builtins, "__import__", mock_import)
    result = index_local(path=str(p), storage_path=storage_dir)
    assert "error" in result
    assert "parquet" in result["error"].lower() or "pyarrow" in result["error"].lower()



"""Tests for JSONL parser and index_local integration."""

import json
import pytest

from jdatamunch_mcp.parser.jsonl_parser import parse_jsonl
from jdatamunch_mcp.tools.index_local import index_local
from jdatamunch_mcp.storage.data_store import DataStore


def test_parse_jsonl_columns(sample_jsonl):
    parsed = parse_jsonl(sample_jsonl)
    assert len(parsed.columns) == 5
    assert [c.name for c in parsed.columns] == ["id", "name", "age", "city", "score"]


def test_parse_jsonl_rows(sample_jsonl):
    parsed = parse_jsonl(sample_jsonl)
    rows = list(parsed.row_iterator)
    assert len(rows) == 10
    assert rows[0][1] == "Alice"
    assert rows[0][0] == "1"


def test_parse_jsonl_null_handling(sample_jsonl):
    parsed = parse_jsonl(sample_jsonl)
    rows = list(parsed.row_iterator)
    frank = rows[5]
    assert frank[2] == ""   # age is None
    assert frank[4] == ""   # score is None


def test_parse_jsonl_sparse_keys(tmp_path):
    """Objects with missing keys should yield empty string for those columns."""
    path = tmp_path / "sparse.jsonl"
    with open(path, "w") as f:
        f.write(json.dumps({"id": 1, "name": "Alice", "extra": "yes"}) + "\n")
        f.write(json.dumps({"id": 2, "name": "Bob"}) + "\n")  # missing "extra"
    parsed = parse_jsonl(str(path))
    rows = list(parsed.row_iterator)
    assert rows[1][2] == ""  # extra is missing → ""


def test_parse_jsonl_skip_bad_lines(tmp_path):
    """Malformed JSON lines are silently skipped."""
    path = tmp_path / "bad.jsonl"
    with open(path, "w") as f:
        f.write(json.dumps({"id": 1, "val": "a"}) + "\n")
        f.write("not valid json\n")
        f.write(json.dumps({"id": 2, "val": "b"}) + "\n")
    parsed = parse_jsonl(str(path))
    rows = list(parsed.row_iterator)
    assert len(rows) == 2


def test_parse_jsonl_empty_file(tmp_path):
    path = tmp_path / "empty.jsonl"
    path.write_text("")
    with pytest.raises(ValueError, match="No valid JSON objects"):
        parse_jsonl(str(path))


def test_parse_jsonl_metadata(sample_jsonl):
    parsed = parse_jsonl(sample_jsonl)
    assert parsed.metadata["estimated_rows"] > 0
    assert parsed.metadata["file_size"] > 0


def test_index_jsonl(sample_jsonl, storage_dir):
    result = index_local(path=sample_jsonl, name="sample-jsonl", storage_path=storage_dir)
    assert "error" not in result
    r = result["result"]
    assert r["rows"] == 10
    assert r["columns"] == 5
    assert r.get("skipped") is not True


def test_index_jsonl_column_types(sample_jsonl, storage_dir):
    index_local(path=sample_jsonl, name="sample-jsonl", storage_path=storage_dir)
    store = DataStore(base_path=storage_dir)
    idx = store.load("sample-jsonl")
    col_types = {c["name"]: c["type"] for c in idx.columns}
    assert col_types["id"] == "integer"
    assert col_types["name"] == "string"
    assert col_types["score"] == "float"


def test_jsonl_nested_objects(tmp_path, storage_dir):
    """Nested dicts become string type."""
    path = tmp_path / "nested.jsonl"
    with open(path, "w") as f:
        for i in range(5):
            f.write(json.dumps({"id": i, "meta": {"key": "val"}}) + "\n")
    result = index_local(path=str(path), name="nested", storage_path=storage_dir)
    assert "error" not in result
    store = DataStore(base_path=storage_dir)
    idx = store.load("nested")
    col_types = {c["name"]: c["type"] for c in idx.columns}
    assert col_types["meta"] == "string"


def test_jsonl_ndjson_extension(tmp_path, storage_dir):
    """.ndjson extension is routed correctly."""
    path = tmp_path / "data.ndjson"
    with open(path, "w") as f:
        for i in range(3):
            f.write(json.dumps({"id": i, "val": str(i)}) + "\n")
    result = index_local(path=str(path), name="ndjson-test", storage_path=storage_dir)
    assert "error" not in result
    assert result["result"]["rows"] == 3



"""Streaming Parquet parser via PyArrow."""

import os
from pathlib import Path
from typing import Generator

from .types import ColumnInfo, ParsedDataset


def _row_generator(path: str, col_count: int) -> Generator:
    """Yield data rows as lists of strings from a Parquet file."""
    import pyarrow.parquet as pq
    pf = pq.ParquetFile(path)
    for batch in pf.iter_batches(batch_size=10_000):
        arrays = [batch.column(i).to_pylist() for i in range(col_count)]
        for row_idx in range(len(batch)):
            yield [
                str(arrays[col][row_idx]) if arrays[col][row_idx] is not None else ""
                for col in range(col_count)
            ]


def parse_parquet(path: str) -> ParsedDataset:
    """Parse a Parquet file and return a streaming ParsedDataset."""
    import pyarrow.parquet as pq

    path = str(Path(path).resolve())
    file_size = os.path.getsize(path)

    pf = pq.ParquetFile(path)
    schema = pf.schema_arrow
    num_rows = pf.metadata.num_rows
    col_count = len(schema)

    columns = [
        ColumnInfo(name=field.name, position=i)
        for i, field in enumerate(schema)
    ]

    metadata = {
        "encoding": "binary/parquet",
        "delimiter": None,
        "header_row": None,
        "estimated_rows": num_rows,
        "file_size": file_size,
        "parquet_num_row_groups": pf.metadata.num_row_groups,
    }

    return ParsedDataset(
        columns=columns,
        row_iterator=_row_generator(path, col_count),
        metadata=metadata,
    )
